import { Collection, InteractionResponse, Message, SnowflakeUtil } from "discord.js";

import { openai } from "./index";
import { formatRequestOrResponse } from "./toOpenAIMessages";
import config from "./config";

export default class Moderation {
  /** Represents cache of messages that have been checked aganist OpenAI moderation API. */
  private static cache = new Collection<string, boolean>();
  private static worker: NodeJS.Timeout | null = null;

  public static async checkMessage(message: Message | InteractionResponse): Promise<boolean> {
    if (this.cache.has(message.id)) {
      return this.cache.get(message.id) as boolean;
    }

    if (message instanceof Message) {
      const warningReaction = message.reactions.resolve("⚠");
      // if bot reacted to that message, we already know that it returned true for moderation API
      if (warningReaction && warningReaction.me) {
        this.cache.set(message.id, true);
        return true;
      }
    }
    
    try {
      const answer = await openai.createModeration({
        input: await formatRequestOrResponse(message),
      });

      const flagged = answer.data.results[0].flagged;
      this.cache.set(message.id, flagged);
      if (flagged) if (message instanceof Message) {
        message.react("⚠");
      }
      else {
        const channelMessage = await message.fetch();
        channelMessage.react("⚠");
      }

      return flagged;
    }
    catch (e) {
      console.error(`Error occured while handling moderation request (${(e as object).constructor.name}):`);
      console.error(e);
      return false;
    }
  }

  public static removeExpiredCacheEntries() {
    const now = Date.now();
    for (const i of this.cache.keys()) {
      if (now - SnowflakeUtil.timestampFrom(i) >= config.limits.time * 2) {
        this.cache.delete(i);
      }
    }
  }
  public static startCacheClearWorker() {
    if (this.worker) return;
    this.worker = setInterval(() => {
      this.removeExpiredCacheEntries();
    },
    1000 * 60 * 15 /* 15 minutes */,
    );
  }

  public static stopCacheClearWorker() {
    if (!this.worker) return;
    clearInterval(this.worker);
    this.worker = null;
  }
}

Moderation.startCacheClearWorker();
